#!/usr/bin/env python
#
# File: calendario.py
# Author: Marco Dalla Stella
# Date: 07/09/2010

import json
from datetime import date, datetime

from sqlalchemy.sql import and_

from waonline.utils.utility import formatData, getGiorniMese
from waonline.apps.gestione.tabelle.tabelle_c  import Voce


# Exceptions varie
class NoDipeTrovato(Exception):
    def __init__(self, msg):
        self.msg = msg

class NoGiornoTrovato(Exception):
    def __init__(self, msg):
        self.msg = msg


class CausaleCal():
    def __init__(self, id, codice, ore, abilitata=True):
        self.id = id
        self.codice = codice
        self.ore = ore
        self.abilitata = abilitata

    def _formatta_ore(self, ore):
        return ('%.2f' % float(ore)).replace('.', ',')

    @property
    def formatta_ore(self):
        if self.ore:
            return self._formatta_ore(self.ore)
        else:
            return ''


class EventoCal():
    def __init__(self, id=None, codice=None, data_inizio=None, data_fine=None, ore=None):
        self.id_evento = id
        self.codice = codice
        self.data_inizio = data_inizio
        self.data_fine = data_fine
        self.ore = ore

    def _format_data(self, data):
        d = ""
        if data:
            d = datetime.strptime(str(data), '%Y%m%d').strftime('%d/%m/%Y')
        return d

    def _object_date(self, data):
        d = None
        if data:
            d = datetime.strptime(str(data), '%Y%m%d')
        return d

    def _formatta_ore(self, ore):
        return ('%.2f' % float(ore)).replace('.', ',')

    @property
    def formatta_ore(self):
        if self.ore:
            return self._formatta_ore(self.ore)
        else:
            return ''

    @property
    def obj_data_inizio(self):
        return self._object_date(self.data_inizio)

    @property
    def obj_data_fine(self):
        return self._object_date(self.data_fine)

    @property
    def _data_inizio(self):
        return self._format_data(self.data_inizio)

    @property
    def _data_fine(self):
        return self._format_data(self.data_fine)

    def _serialize(self):
        return {'evento': { 'id_evento': self.id_evento, 'codice_evento': self.codice,
                            'data_inizio': self._data_inizio, 'data_fine': self._data_fine, 'ore_evento': self.ore}}


class VoceCal():
    def __init__(self, id=None, prefisso=None, voce=None, suffisso=None, dato_base=None, costo_unitario=None, perc=None, importo=None):
        self.id_voce = id
        self.prefisso = prefisso
        self.voce = voce
        self.suffisso = suffisso
        self.dato_base = dato_base
        self.costo_unitario = costo_unitario
        self.perc = perc
        self.importo = importo

    def _serialize(self):
        return {'voce': {'id_voce': self.id_voce, 'prefisso': self.prefisso, 'voce': self.voce, 'suffisso': self.suffisso,
                         'dato_base': self.dato_base, 'costo_unitario': self.costo_unitario, 'perc': self.perc, 'importo': self.importo}}


class LavoroDomCal():
    def __init__(self, id=None, data_consegna=None, ora_consegna=None, data_riconsegna=None, ora_riconsegna=None, descrizione=None, quantita=None):
        self.id = id
        self.data_consegna = data_consegna
        self.ora_consegna = ora_consegna
        self.data_riconsegna = data_riconsegna
        self.ora_riconsegna = ora_riconsegna
        self.descrizione = descrizione
        self.quantita = quantita

    def _format_data(self, data):
        d = ""
        if data:
            d = datetime.strptime(str(data), '%Y%m%d').strftime('%d/%m/%Y')
        return d

    def _format_ora(self, ora):
        d = ""
        if ora:
            d = datetime.strptime(str(ora), '%H.%M').strftime('%H:%M')
        return d

    @property
    def _data_consegna(self):
        return self._format_data(self.data_consegna)

    @property
    def _data_riconsegna(self):
        return self._format_data(self.data_riconsegna)

    @property
    def _ora_consegna(self):
        return self._format_ora(self.ora_consegna)

    @property
    def _ora_riconsegna(self):
        return self._format_ora(self.ora_riconsegna)

    def _serialize(self):
        return {'lavdom': {'id_lavdom': self.id, 'data_consegna': self._data_consegna, 'ora_consegna': self._ora_consegna,
                           'data_riconsegna': self._data_riconsegna, 'ora_riconsegna': self._ora_riconsegna, 'descrizione': self.descrizione,
                           'quantita': self.quantita}}



class Giorno():
    def __init__(self, dati_giorno, causali_no_cal=False):
        self.id = dati_giorno.id
        self.giorno = dati_giorno.giorno_mese
        self.ore_teoriche = dati_giorno.ore_teoriche
        self.ore_lavorate = dati_giorno.ore_lavorate
        self.ore_retribuite = dati_giorno.ore_retribuite
        self.ecc = dati_giorno.ecc
        self.abilitato = True
        self.causali_giorno = []

        if not causali_no_cal:
            causali_no_cal = []

        for causale in dati_giorno.calendarioCaus_rel:
            c = CausaleCal(id=causale.id, codice=causale.codice, ore=causale.ore)
            if c.codice in causali_no_cal:
                c.abilitata = False
            self.causali_giorno.append(c)

    def _formatta_ore(self, ore):
        return ('%.2f' % float(ore)).replace('.', ',')

    @property
    def formatta_ore_teoriche(self):
        if self.ore_teoriche:
            return self._formatta_ore(self.ore_teoriche)
        else:
            return ''

    @property
    def formatta_ore_lavorate(self):
        if self.ore_lavorate:
            return self._formatta_ore(self.ore_lavorate)
        else:
            return ''

    def ricava_causale_giorno(self, numero_causale):
        causale = None
        try:
            causale = self.causali_giorno[numero_causale]
        except IndexError:
            causale = None
        return causale


    def __unicode__(self):
        return u"<Giorno %s ore t/l/r: %s %s %s>" % (
            self.giorno, self.ore_teoriche, self.ore_lavorate, self.ore_retribuite)

    def __repr__(self):
        return unicode(self).encode('utf-8')

class PresenzaCal():
    def __init__(self, presenza, causali_no_cal=False):
        self.id = presenza.id
        self.id_dipe = presenza.dipendente_id
        self.periodo = presenza.periodo
        self.codice = presenza.dipendente_rel.codice
        self.matricola = presenza.dipendente_rel.matricola
        self.nominativo = presenza.dipendente_rel.nominativo
        self.is_collab = presenza.dipendente_rel.is_collab
        self.mansione = presenza.mansione
        self.data_cess = 0
        self.data_ass = presenza.dipendente_rel.data_ass
        if presenza.dipendente_rel.data_cess_fut:
            self.data_cess = presenza.dipendente_rel.data_cess_fut
        elif presenza.dipendente_rel.data_cess:
            self.data_cess = presenza.dipendente_rel.data_cess
        self.qualifica = presenza.qualifica
        self.livello = presenza.livello
        self.tipo_pm = presenza.tipo_pm
        self.perc_pm = presenza.perc_pm
        self.pos_inail = presenza.pos_inail
        self.centro_costo = presenza.centro_costo
        self.evento_mp = presenza.evento_mp

        self.giorni = []
        self.eventi = []
        self.voci = []
        self.lavdom = []
        self.note = presenza.note

#        presenze primo giorno utile OT ed RS per ricalcolo ore retribuite
        self.pgu_ore_teoriche = 0
        self.pgu_ore_retribuite = 0
        self.pgu_ore_retribuite_rs = 0
        try:
            pgu = [g for g in presenza.calendario_rel if g.ecc == None and g.ore_teoriche > 0][0]
            self.pgu_ore_teoriche = pgu.ore_teoriche
            self.pgu_ore_retribuite = pgu.ore_retribuite
        except:pass
        try:
            pgu = [g for g in presenza.calendario_rel if g.ecc == "RS"][0]
            self.pgu_ore_retribuite_rs = pgu.ore_retribuite
        except:pass

#       Carico le causali dei giorni calendario
        for dati_giorno in presenza.calendario_rel:
            self.giorni.append(Giorno(dati_giorno, causali_no_cal))

#       Carico gli eventi calendario
        for dati_eventi in presenza.eventi_rel:
            e = EventoCal(dati_eventi.id, dati_eventi.codice, dati_eventi.data_inizio, dati_eventi.data_fine, dati_eventi.ore)
            self.eventi.append(e)

#       Carico le voci calendario
        for dati_voce in presenza.voci_rel:
            v = VoceCal(dati_voce.id, dati_voce.prefisso, dati_voce.voce,
                     dati_voce.suffisso, dati_voce.dato_base, dati_voce.costo_unitario,
                     dati_voce.perc, dati_voce.importo)
            self.voci.append(v)

#       Carico i dati lavoro domicilio
        for dati_lavdom in presenza.lavori_dom_rel:
            e = LavoroDomCal(dati_lavdom.id, dati_lavdom.data_consegna, dati_lavdom.ora_consegna, dati_lavdom.data_riconsegna,
                             dati_lavdom.ora_riconsegna, dati_lavdom.descrizione, dati_lavdom.quantita)
            self.lavdom.append(e)

    def _ricava_giorni(self, dal=None, al=None):
        if not dal:
            dal = self.giorni[0].giorno

        if not al:
            al = self.giorni[-1].giorno

        giorni = []
        for giorno in self.giorni:
            if giorno.giorno >= dal and giorno.giorno <= al:
                giorni.append(giorno)
        return giorni

    def __unicode__(self):
        return u"<PresenzaCal: %s %s>" % (self.nominativo, self.periodo)

    def __repr__(self):
        return unicode(self).encode('utf-8')


class Calendario():
    def __init__(self, ditta_id, periodo, causali_no_calendario):
        self.ditta_id = ditta_id
        self.id_dipendenti = []
        self.presenze = {}
        self.periodo = str(periodo)
        self.periodo_anno = int(self.periodo[0:4])
        self.periodo_mese = int(self.periodo[4:6])
        self.causali_no_calendario = causali_no_calendario

    def aggiungi_presenze_dipe(self, presenze_dipe):
        if not presenze_dipe.dipendente_id: # salta i dipendenti eliminati dall'anagrafica
            return
        if presenze_dipe.dipendente_id not in self.presenze:
            self.presenze[presenze_dipe.dipendente_id] = PresenzaCal(presenze_dipe, self.causali_no_calendario)
            self.id_dipendenti.append(presenze_dipe.dipendente_id)

    def ricava_id_dipendenti(self):
        return self.id_dipendenti

    def ricava_presenze(self, id_dipe=[]):
        if not id_dipe:
            id_dipe = self.id_dipendenti

        presenze = []
        for dipe in id_dipe:
            presenze.append(self.ricava_presenza(dipe))
        return presenze

    def ricava_giorno_dipe(self, id_dipe, num_giorno):
        if not id_dipe in self.presenze:
            raise NoDipeTrovato('Dipendente con id %s non trovato.' % id_dipe)

        try:
            return self.presenze[id_dipe].giorni[num_giorno - 1]
        except IndexError:
            raise NoGiornoTrovato('Il giorno %d non e\' stato trovato nel calendario.' % num_giorno)

    def ricava_presenza(self, id_dipe):
        if not id_dipe in self.presenze:
            raise NoDipeTrovato('Dipendente con id %s non trovato.' % id_dipe)

        presenza = self.presenze[id_dipe]
        return presenza

    def ricava_giorni_dipe(self, id_dipe, dal=None, al=None):
        presenza = self.ricava_presenza(id_dipe)
        giorni = []
        for giorno in presenza._ricava_giorni(dal, al):
            giorni.append(giorno)
        return giorni

    def ricava_eventi_dipe(self, id_dipe):
        if not id_dipe in self.presenze:
            raise NoDipeTrovato('Dipendente con id %s non trovato.' % id_dipe)
        return self.presenze[id_dipe].eventi

    def ricava_evento_dipe(self, id_dipe, id_riga):
        if not id_dipe in self.presenze:
            raise NoDipeTrovato('Dipendente con id %s non trovato.' % id_dipe)

        try:
            evento = self.presenze[id_dipe].eventi[id_riga]
        except IndexError:
            evento = None

        return evento

    def ricava_voci_dipe(self, id_dipe):
        if not id_dipe in self.presenze:
            raise NoDipeTrovato('Dipendente con id %s non trovato.' % id_dipe)
        return self.presenze[id_dipe].voci

    def ricava_lavdom_dipe(self, id_dipe):
        if not id_dipe in self.presenze:
            raise NoDipeTrovato('Dipendente con id %s non trovato.' % id_dipe)
        return self.presenze[id_dipe].lavdom


    def __unicode__(self):
    #return u"<Calendario: %s %s %s>" % (self.ditta_id, self.periodo, self.presenze)
        return u"<Calendario: %s %s>" % (self.ditta_id, self.periodo)

    def __repr__(self):
        return unicode(self).encode('utf-8')

class GeneraDatiCal():
    def __init__(self, rbase, calendario):
        self.rbase = rbase
        self.calendario = calendario
        self.CAUSALI_MAX = 5
        self.SETTIMANA = ["Lu", "Ma", "Me", "Gi", "Ve", "Sa", "Do"]
        self.giorni_in_forza = None

    def _correggi_id_dipe(self, id_dipe):
        if not id_dipe:
            id_dipe = [int(x) for x in self.calendario.id_dipendenti]
        else:
            if not isinstance(id_dipe, list):
                id_dipe = [int(id_dipe)]
            else:
                id_dipe = [int(x) for x in id_dipe]
        return id_dipe

    def genera_dati_testata(self, id_dipe, type='json'):
        id_dipe = self._correggi_id_dipe(id_dipe)

        anagrafiche = []
        for dipe in id_dipe:
            presenza = self.calendario.ricava_presenza(dipe)
            anagrafiche.append({'dipendente': {'codice': presenza.codice, 'nominativo': presenza.nominativo,
                                               'data_ass': presenza.data_ass, 'qualifica': presenza.qualifica,
                                               'livello': presenza.livello, 'centro_costo': presenza.centro_costo}})
        return self._return_data(anagrafiche, 'dipendenti', type)

    def genera_eventi(self, id_dipe=None, type='json'):
        id_dipe = self._correggi_id_dipe(id_dipe)

        eventi = []
        for dipe in id_dipe:
            i = 0
            for evento in self.calendario.ricava_eventi_dipe(dipe):
                e = evento._serialize()
                if evento.codice in self.rbase.session['causali_db']:
                    e['evento']['descrizione'] = self.rbase.session['causali_db'][evento.codice]['descrizione']
                else:
                    e['evento']['descrizione'] = ""
                e['evento']['id_riga'] = i
                eventi.append(e)
                i += 1

        # Inserisco un evento vuoto
        e = EventoCal()._serialize()
        e['evento']['id_riga'] = i
        eventi.append(e)

        return self._return_data(eventi, 'eventi', type)

    def genera_evento_id(self, id_dipe=None, id=0, type='json'):
        id_dipe = self._correggi_id_dipe(id_dipe)

        eventi = []
        for dipe in id_dipe:
            evento = self.calendario.ricava_evento_dipe(dipe, id)
            if evento:
                e = evento._serialize()
                if evento.codice in self.rbase.session['causali_db']:
                    e['evento']['descrizione'] = self.rbase.session['causali_db'][evento.codice]['descrizione']
                else:
                    e['evento']['descrizione'] = ""
                e['evento']['id_riga'] = id
                eventi.append(e)

        return self._return_data(eventi, 'eventi', type)

    def genera_voci(self, id_dipe=None, type='json'):
        id_dipe = self._correggi_id_dipe(id_dipe)

        voci = []
        for dipe in id_dipe:
            i = 0
            for voce in self.calendario.ricava_voci_dipe(dipe):
                descrizione_voce = self.rbase._q(Voce.descrizione)\
                                             .filter(and_(Voce.codice == voce.voce,
                                                          Voce.periodo_inizio <= self.calendario.periodo,
                                                          Voce.periodo_fine >= self.calendario.periodo,
                                                          )
                                                     ).first()[0]
                if descrizione_voce:
                    descrizione_voce = descrizione_voce[0]

                v = voce._serialize()
                v['voce']['descrizione'] = descrizione_voce
                v['voce']['id_riga'] = i
                voci.append(v)
                i+=1

        # Inserisco una voce vuota
            v = VoceCal()._serialize()
            v['voce']['id_riga'] = i
            voci.append(v)

        return self._return_data(voci, 'voci', type)

    def genera_note(self, id_dipe=[], type='json'):
        id_dipe = self._correggi_id_dipe(id_dipe)

        righe = []
        for dipe in id_dipe:
            presenza = self.calendario.ricava_presenza(dipe)
            righe.append({'nota': {'nota': presenza.note}})

        return self._return_data(righe, 'note', type)

    def genera_lavdom(self, id_dipe=None, type='json'):
        id_dipe = self._correggi_id_dipe(id_dipe)

        lavdom = []
        for dipe in id_dipe:
            i = 0
            for ldom in self.calendario.ricava_lavdom_dipe(dipe):
                l = ldom._serialize()
                l['lavdom']['id_riga'] = i
                lavdom.append(l)
                i += 1

        # Inserisco un evento vuoto
        l = LavoroDomCal()._serialize()
        l['lavdom']['id_riga'] = i
        lavdom.append(l)

        return self._return_data(lavdom, 'lavdoms', type)

    def genera_calendario(self, id_dipe=[], dal=None, al=None, type='json'):
        id_dipe = self._correggi_id_dipe(id_dipe)

        self.id_cella = 0
        righe = []

        for dipe in id_dipe:
            try:
                presenza = self.calendario.ricava_presenza(dipe)
                righe += self._genera_data_presenza(presenza, dal, al)
            except:
                pass

        return self._return_data(righe, 'calendario', type)

    def _return_data(self, data, nome, type):
        if type == 'xml':
            o = ObjectToXML()
            return o.getXML(data, nome)
        else:
            return json.dumps(data)

    def _genera_data_presenza(self, presenza, dal, al):
        da_giorno = 1
        a_giorno = getGiorniMese(self.calendario.periodo)
        if formatData(presenza.data_ass, "int", "%Y%m") == presenza.periodo:
            da_giorno = formatData(presenza.data_ass, "int", "%d")
        if presenza.data_cess and formatData(presenza.data_cess, "int", "%Y%m") == presenza.periodo:
            a_giorno = formatData(presenza.data_cess, "int", "%d")
        self.giorni_in_forza = range(da_giorno, a_giorno + 1)

        righe = []
        for giorno in presenza._ricava_giorni(dal, al):
            righe += self._genera_data_giorno(presenza.id_dipe, presenza.codice + '<br />' + presenza.nominativo, giorno)
        return righe

    def _genera_data_giorno(self, id_dipe, nome, giorno):
        day = date(self.calendario.periodo_anno, self.calendario.periodo_mese, giorno.giorno)
        col_desc = '%s %02d' % (self.SETTIMANA[day.weekday()], day.day)

######### Riga 1
        causale = 'FG' if giorno.ecc == 'FG' else 'OT'
        ore = giorno.ecc if giorno.ecc in ['RS', '--'] else giorno.formatta_ore_teoriche
        value = causale + ' ' + ore if ore.isdigit else ore
        cau_riga_1 = {'rows': 'C0', 'nome': nome, 'id_dipe': id_dipe, 'value': value, 'cols': col_desc}
        if giorno.giorno not in self.giorni_in_forza:
            cau_riga_1['enabled'] = False
######### Riga 2
        ore = '' if giorno.ecc in ['RS', '--'] or giorno.giorno not in self.giorni_in_forza else giorno.formatta_ore_lavorate
        value = 'OL ' + ore
        cau_riga_2 = {'enabled': False, 'rows': 'C1', 'nome': nome, 'id_dipe': id_dipe, 'value': value, 'cols': col_desc}

######### Riga 3, 4, 5
        i = 1
        righe_causali = []
        if giorno.giorno not in self.giorni_in_forza:
            i += 1
            row = 'C' + str(i)
            ore = ''
            if giorno.giorno < self.giorni_in_forza[0]:
                value = 'AA' + ' ' + ore
            else:
                value = 'ZZ' + ' ' + ore
            cau_riga_x = {'rows': row, 'nome': nome, 'id_dipe': id_dipe,
                          'value': value, 'cols': col_desc,
                          'enabled': False}
            righe_causali.append(cau_riga_x)
        else:
            for causale in giorno.causali_giorno:
                i += 1
                row = 'C' + str(i)
                ore = causale.formatta_ore
                value = causale.codice + ' ' + ore
                cau_riga_x = {'rows': row, 'nome': nome, 'id_dipe': id_dipe,
                              'value': value, 'cols': col_desc,
                              'enabled': causale.abilitata}
                righe_causali.append(cau_riga_x)

        while i < (self.CAUSALI_MAX + 1):
            i += 1
            row = 'C' + str(i)
            cau_riga_x = {'rows': row, 'nome': nome, 'id_dipe': id_dipe, 'value': '', 'cols': col_desc}
            if giorno.giorno not in self.giorni_in_forza:
                cau_riga_x['enabled'] = False
            righe_causali.append(cau_riga_x)

        return [cau_riga_1, cau_riga_2] + righe_causali


#Classe ObjectToXML per trasformare da struttura Python a XML
class ObjectToXML():
    def __init__(self):
        self.adapt = {
            dict: self.getXML_dict,
            list: self.getXML_list,
            tuple: self.getXML_list,
        }

    def getXML(self, obj, objname):
        if obj == None:
            return ""
        if self.adapt.has_key(obj.__class__):
            return self.adapt[obj.__class__](obj, objname)
        else:
            return "<%(n)s>%(o)s</%(n)s>"%{'n':objname,'o':str(obj)}

    def getXML_dict(self, indict, objname=None):
        h = "<%s>"%objname
        for k, v in indict.items():
            h += self.getXML(v, k)
        h += "</%s>"%objname
        return h

    def getXML_list(self, inlist, objname=None):
        h = ""
        for i in inlist:
            h += self.getXML(i, objname)
        return h
