# coding: latin-1

"""
===========================
    Rbase (modulo di base)
===========================

.. autoclass
   :special-members:

:File: rbase.py
:Author: Luca Zaccaria luca.z@essepaghe.it
:Date: 04/06/2010
"""
import inspect
import re
import os
from datetime import datetime
from urllib2 import unquote

from waonline.core.wadecorators import combomethod
from waonline.core import waologgin as wl
from waonline.utils.utility import Risposta, formatData
from waonline.core.admin.tabelle.admin_log import AdminLog

DOCTYPES = {"xml": '<?xml version="1.0" encoding="UTF-8"?>', "xhtml": \
        '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" \
        http: //www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n\n', "html": \
        '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" \
        "http: //www.w3.org/TR/html4/strict.dtd">\n\n'}


class RBase(object):
    """:class:`RBase` fornisce l'interfaccia di base per l'interazione tra
    :mod:`wademon`, :mod:`utenti` sqlalchemy etc.

    Molti dei metodi o delle propriet sono resi privati (_ o __) per eveitare
    l'eventuale esecuzione diretta tramite richieste html.

    .. data:: sessiondb

       Connessione **sqlalchemy** al database.

    .. data:: session

       Oggetto :class:`Session` contenente la sessione http di wademon.

    .. data:: args

       Argomenti passati dal client alla request di wadaemon.

    .. data:: _files

       Oggetto di wademon contenente i file passati per l'upload.

    .. data:: _task

       task da eseguire richiesto da wadaemon che RBase recupera all'init dalla
       request.

    .. data:: risposta

       oggetto :class:`Risposta` contenente tutte le informazioni che wademon
       utilizzer per la risposta alla chiamata del client.

    .. data:: user

       oggetto :class:`Utente` con l'utente attuale.

    .. data:: _dipti_id

       Lista di id dei dipendenti (se presenti) o [] (lista vuota) se nessuno

    .. data:: _cod_ditta

       stringa di 7 caratteri con il codice della ditta attuale, RBase fa un
       zfill sul codice aggiungedo automaticamente gli 0 all'inizio della
       stringa, se necessario.

    .. data:: _periodo

       intero con il periodo attuale in formato YYYYMM.
    """
    __user_ditteid__ = None
    __ditta__ = None
    __dipti__ = None
    _cod_ditta = None
    _periodo = None
    _dipti_id = None
    _to_log = []
    _classi = []
    codice_studio = None
    @wl.timewarn
    def __init__(self, sessiondb, session=None, request=None):
        if session == None:
            plist = ["sessiondb", "session", "request", "__user_ditteid__",
                    "__ditta__", "__dipti__", "_cod_ditta", "user", "_files",
                    "_path", "_headers", "_force_response", "_periodo",
                    "_dipti_id", "_to_log", "codice_studio", "args",
                    "utenti_manager", "db_manager"]
            for prop in plist:
                try:
                    setattr(self, prop, getattr(sessiondb, prop))
                except:
                    wl.errore("Errore assegnando %s" % prop)
        else:
            self.sessiondb = sessiondb
            self.session = session
            self.request = request
            self.args = self.request.arguments
            self._files = self.request.files
            self._path = self.request.path
            self._headers = self.request.headers
            self._um = self.utenti_manager = self.request.utenti_manager
            self.db_manager = self.request.db_manager
            self._force_response = False
            for nome, valore in self.args.items():
                if isinstance(valore, (list, tuple)):
                    nlist = []
                    for l in valore:
                        if isinstance(l, str):
                            nlist.append(unquote(l))
                        else:
                            nlist.append(l)
                    if len(valore) == 1:
                        self.args[nome] = valore[0]
                else:
                    self.args[nome] = unquote(valore)
            #: Spostato in wademon.py il caricamento dell'utente
            self.user = self.request.user
            self._getDefaultVars()
            if self._cod_ditta:
                self.info('Codice ditta: %s' % self._cod_ditta)
        self._task = self.getArg('task', default='_noTask', nolist=True)
        self.risposta = Risposta()
        self._post_init_()

    @combomethod
    def extend(self, newCls):
        for nome, valore in vars(newCls).iteritems():
            if not nome.startswith('__'):
                setattr(self, nome, valore)
                for classe in self._classi:
                    setattr(classe, nome, valore)

    def _add_to_log(self, funzione=None, modulo=None):
        """ Aggiunge Informazioni al log nel database """
        fr = inspect.stack()
        mo = inspect.getframeinfo(fr[1][0])
        modulo = modulo or os.path.basename(mo.filename).replace(".py","")
        funzione = funzione or mo.function
        if not isinstance(funzione, list):
            funzione = [funzione]
        for fn in funzione:
            m = fn.split(".")
            if len(m) == 2:
                nmodulo = m[0]
                nfunzione = m[1]
            else:
                nmodulo = modulo
                nfunzione = m[0]
            self._to_log.append(nmodulo + "." + nfunzione)

    @property
    def _cod_studio(self):
        if self.user.studio_codice:
            try:
                return self.user.studio_codice.zfill(3)
            except:
                try:
                    return self.user.studio_codice
                except:
                    pass
        return None

    def _getDefaultVars(self):
        """ Carica gli attributi di default """
        self._cod_ditta = self.ga("cod_ditta")
        self._dipti_id = self.ga('id_dipte', default=[], forcelist=True)
        self._periodo = self.ga('periodo')
        if self._periodo:
            self._periodo = formatData(self._periodo, "int", "%Y%m")
            if self._periodo:
                if not re.match(r'^\d{4}\d{2}$', str(self._periodo)):
                    self._periodo = None
        if self._cod_ditta:
            self._cod_ditta = str(self._cod_ditta).zfill(7).upper()
        if self._dipti_id:
            try:
                if self._dipti_id[0].find(",") != -1:
                    self._dipti_id = self._dipti_id[0].split(",")
            except:
                pass
            for i in range(len(self._dipti_id)):
                try:
                    self._dipti_id[i] = int(self._dipti_id[i])
                except Exception, e:
                    self._dipti_id[i] = None
            for e in self._dipti_id:
                if e == None:
                    self._dipti_id.remove(e)

    @property
    def _impostazioni(self):
        """ Dizionario con le impostazioni della ditta corrente """
        if self._cod_ditta:
            return self.user.ditta_by_codice(
                    self._cod_ditta).get_all_impostazioni()
        else:
            return self.user.get_all_impostazioni()

    @property
    def _preferenze(self):
        """ :class:`wadecohelpers.EteDumbObj` con le preferenze della ditta
        corrente """
        if self._cod_ditta:
            return self.user.ditta_by_codice(self._cod_ditta).get_all_prefs()
        else:
            return self.user.get_all_prefs()

    @property
    def _preflst(self):
        """ lista di tuple con id, nome e valore della preferenza per la ditta
            corrente"""
        if self._cod_ditta:
            return self.user.ditta_by_codice(self._cod_ditta).get_all_prefs(True)
        else:
            return self.user.get_all_prefs(True)

    @property
    def _dipti_count(self):
        """ Numero di id dipendenti in :data:`_dipti_id` """
        return len(self._dipti_id)

    @property
    def _dipte_id(self):
        """ Id del primo dipendente della lista di id dei dipendenti se
        presenti o 0 se non c' nulla.
        """
        if (self._dipti_id):
            return self._dipti_id[0]
        else:
            return 0

    def _fresponse(self, force="json"):
        """ Forza il formato di risposta il formato pu essere "json" o "xml",
        default "json"
        """
        self._force_response = force

    def _ftipoDati(self, force='string'):
        """ Forza il tipo di dati per la risposta, default "string"
        """
        self.risposta.tipoDati = force

    def _post_init_(self):
        """ Permette di eseguire ogni operazione necessaria subito dopo l'init
        della classe.
        E' possibile definire nel :func:`_post_init_` di ogni classe derivata
        una qualsiasi serie di operazione che verranno eseguite subito dopo
        l'inizializzazione della classe. Un'alternativa a
        super(classe, slef).__init__()
        """
        pass

    def _pre_task_(self):
        """ Permette di esguire ogni operazione necessaria subito prima del
        task
        """
        pass

    def errore(self, messaggio, inline=1, dept=2):
        """ Inserisce un messaggio d'errore nel file di log; *inline* a True
        forza il messaggio su di una singola riga.
        """
        wl.errore(messaggio, inline=inline, dept=dept)

    def info(self, messaggio, inline=1, dept=2):
        """ Inserisce un messaggio nel file di log; *inline* a True
        forza il messaggio su di una singola riga.
        """
        wl.info(messaggio, inline=inline, dept=dept)

    def debug(self, messaggio, inline=1, dept=2):
        """ Inserisce un messaggio di debug nel file di log; *inline* a True
        forza il messaggio su di una singola riga.
        """
        wl.debug(messaggio, inline=inline, dept=dept)

    def warn(messaggio, inline=1, dept=2):
        """ Inserisce un messaggio di debug nel file di log; *inline* a True
        forza il messaggio su di una singola riga.
        """
        wl.warn(messaggio, inline=inline, dept=dept)

    def _setDocType(self, doctype="xml"):
        """ Forza il doctype di risposta.contenuto "html" o "xml",
        default "xml"
        """
        if self.risposta.contenuto:
            self.risposta.contenuto = DOCTYPES.get(
                doctype, DOCTYPES[doctype]) + self.risposta.contenuto
        else:
            self.risposta.contenuto = DOCTYPES.get(doctype, DOCTYPES[doctype])

    def _pop(self, nomi):
        """ Effettua un pop su self.args e ne restituisce il risultato; *nomi*
        pu essere sia una chiave singola che una lista di chiavi
        """
        risultati = []
        if not isinstance(nomi, list):
            try:
                return risultati.append(self.args.pop(nomi))
            except:
                return None
        for nome in nomi:
            try:
                risultati.append(self.args.pop(nome))
            except:
                pass
        return risultati

    def getArg(self, nome, default=None, nolist=False, forcelist=False):
        """ Ritorna il valore di una variabile contenuta nella request;
        *nome* il nome della variabile da recuperare nella request, *default*
        imposta il valore di default in caso non trovasse la variabile,
        *nolist* in caso la variabile contenesse una lista ne ritorna solo
        il primo elemento, *forcelist* forza il risultato ad una lista in caso
        non lo fosse.

        .. note::
           E' consigniabile utilizzare il decoratore :class:`wadecorators.
           WaAuth` per i metodi in modo da automatizzare il controllo sulle
           variabili e renderle disponibili come argomenti del metodo.
        """
        if nome in self.args:
            value = self.args[nome]
        else:
            value = default
        if nolist and type(value) in [list, tuple]:
            if value:
                value = value[0]
        elif forcelist and type(value) not in [list, tuple]:
            value = [value]
        try:
            if value.lower() == 'true':
                value = True
            elif value.lower() == 'false':
                value = False
            elif value.lower() == 'null':
                value = None
        except:
            pass
        return value
    #: Alias per :func:`getArg`
    ga = getArg
    def _noTask(self):
        """ Medoto da eseguire in caso di mancato task """
        self._addErr('Nessun task da eseguire')

    def _log_func(self, user_id, cod_ditta, periodo, tipo, task, data_inizio,
            data_fine, inp, outp):
        """ logga la funzione indicata """
        try:
            log = AdminLog(
                user_id = user_id,
                ip = self.request.remote_ip,
                request = self.request.uri,
                cod_ditta = cod_ditta,
                periodo = periodo,
                data_inizio = data_inizio,
                data_fine = data_fine,
                mtype = tipo,
                task = task,
                minput = inp,
                output = outp,
            )
            self.sessiondb.add(log)
            self.sessiondb.commit()
        except Exception, err:
            self.errore("Errore nel logdb:%s" % str(err))
            self.sessiondb.rollback()

    def doTask(self, *args, **kwargs):
        """ Si occupa di eseguire il task richiesto da wademon.

        .. danger:: **NON TOCCARE**

        .. attention:: rinominare in _doTask
        """
        self._pre_task_()
        if self.user:
            self.debug("doTask: user: %s %s" % (self.user, self.user.sessiondb))
        else:
            self.debug("doTask: Utente a None")

        #: tento di recuperare il metodo da chiamare e assegnarlo a task
        if 'method' in kwargs:
            self._task = kwargs.get('method', self._task)
            del kwargs['method']
        if hasattr(self, self._task):
            data_inizio = datetime.now()
            if self.user:
                user_id = self.user.id
            else:
                user_id = None
            try:
                getattr(self, self._task)(*args, **kwargs)
            except:
                try:
                    self.sessiondb.rollback()
                except:
                    pass
                raise
            finally:
                self.sessiondb.rollback()
                if self.ga('type', 'app') + "." + self._task in self._to_log:
                    data_fine = datetime.now()
                    try:
                        del self.args['transaction']
                    except:
                        pass
                    self._log_func(user_id=user_id, cod_ditta=self._cod_ditta,
                        periodo=self._periodo, tipo=self.ga('type'),
                        task=self._task, data_inizio=data_inizio,
                        data_fine=data_fine, inp=unicode(self.args),
                        outp=unicode(self.risposta.contenuto)
                    )
                    self.sessiondb.rollback()
        else:
            self._addErr("NON trovato; %s" % self._task)
        from waonline.utils.utility import toJSONResponse, toXMLResponse
        if (self.risposta.errori
            or self.risposta.messaggi
            or self._force_response == 'json'):
            self._repHead("Content-Type", "text/plain")
            self.risposta = toJSONResponse(self.risposta)
        elif self._force_response == "xml":
            self._repHead("Content-Type", "text/xml")
            self.risposta = toXMLResponse(self.risposta)

        try:
            self.sessiondb.commit()
            self.sessiondb.close()
            del self.sessiondb
        except Exception, err:
            self.warn("ERRORE disconnessione: %s" % err)
            pass

        return self.risposta

    def _addCookie(self, name, value, domain=None, expires=None, path="/",
            expires_days=None, **kwargs):
        """ Aggiunge un cookie alla lista """
        self.risposta.cookies.append(dict(name=name, value=value,
            domain=domain, expires=expires, path=path,
            expires_days=expires_days, kwargs=kwargs))

    def _addErr(self, messaggio, coords=None):
        """ Aggiunge un *messaggio* d'errpre nella coda messaggi della risposta
        e anche, a scopo di debug, al file di log. In messaggi in coda verranno
        visualizzati poi nella finestra messaggi del client.

        E' possibile specificare in *coords* un dizionario con le coordinate da
        utilizzare per settare il fuoco sella griglia del calendario.

        .. note:: In realt *coords* potrebbe essere un qualsiasi dizionario da
           far utilizzare a delle funzioni di callback.
        """
        self.errore("Err: %s" % messaggio)
        self.risposta.messaggi.append(
                {'tipo': 'errore', 'msg': messaggio, 'coords': coords})

    def _addMess(self, messaggio, coords=None):
        """ Aggiunge un *messaggio* nella coda messaggi della risposta vedi
        :func:`_addErr`
        """
        self.info("Msg: %s" % messaggio)
        self.risposta.messaggi.append(
                {'tipo': 'messaggio', 'msg': messaggio, 'coords': coords})

    def _addStato(self, messaggio, timeout=None):
        """ Aggiunge un *messaggio* di stato alla coda dei messaggi della
        risopsta. E' possibile specificare un *timeout* in millisecondi per il
        messaggio dopo il quale verr cancellato nel client.

        .. note:: Solo l'ultimo messaggio di stato verr visualizzato nel client
        """
        self.risposta.messaggi.append(
                {'tipo': 'stato', 'msg': messaggio, 'timeout': timeout})

    def _addConferma(self, messaggio, bottoni="sn"):
        """ Aggiunge un *messaggio* di conferma alla coda dei messaggi della
        risposta e i *bottoni* da visualizzare nella finestra di dialogo.
        I bottoni utilizzabili sono **s**\i, **n**\o, **c**\onferma,
        **a**\\nnulla
        """
        self.risposta.messaggi.append(
                {'tipo': 'conferma', 'msg': messaggio, 'tasti': bottoni})
    #: Alias per :meth:`_addConferma`
    _addConfermacalendario = _addConferma

    def  _hasErrori(self):
        """ Restituisce True se la coda messaggi della risposta contiene
        degli errori
        """
        return len([l for l in self.risposta.messaggi \
                 if l['tipo'] == 'errore']) != 0

    def  _hasConferma(self):
        """ Restituisce True se la coda messaggi della risposta contiene
        degli una rischiesta di conferma
        """
        return len([l for l in self.risposta.messaggi \
                 if l['tipo'] == 'conferma']) != 0

    def _addHead(self, nome, valore):
        """ Aggiunge un Header alla risposta """
        self.risposta.header.append((nome, valore))

    def _repHead(self, nome, valore):
        """ Sostituisce un header nella risposta """
        self._remHead(nome)
        self._addHead(nome, valore)

    def _remHead(self, nome):
        """ Rimuove un header dalla risposta """
        try:
            self.risposta.header.remove(
                self.risposta.header[
                    [l[0].lower() for l in self.risposta.header].index(
                        nome.lower())])
        except:
            pass

    def _noNone(self, valore):
        """ Ritorna una stringa unicode per *valore*, una stringa vuota
        in caso il valore sia None
        """
        if valore == None:
            return u''
        return unicode(valore)
    #: Alias per :func:`_noNone`
    _nn = _noNone

    def _query(self, query, *qfilters, **kwargs ):
        """ Wrapper per self.sessiondb.query con supporto filtri multiplo che
        vuol dire che la query **_query(Ditta).filter(Ditta.codice==codice)**
        pu essere pi comodamente scritta **_query(Ditta,
        Ditta.codice==codice)**
        """
        if query.__class__.__name__ == 'InstrumentedAttribute':
            query = [query]
            query.extend(qfilters)
            query = tuple(query)
            q = self.sessiondb.query(*query, **kwargs)
        else:
            q = self.sessiondb.query(query, **kwargs)
            for qfilter in qfilters:
                q = q.filter(qfilter)
        return q
    #: Alias per :func:`_query`
    _q = _query

    def _ex(self, query, commit=False):
        """ Wrapper for self.sessiondb.execute """
        q = self.sessiondb.execute(query)
        if commit:
            self.sessiondb.commit()
        return q
