# -*- coding: utf-8 -*-
"""
.. _tbl-user:

Modulo per gli utenti del database (nuova versione)
===================================================

.. index:: NewUtente, NewUtentePref

"""
__fine_name__ = "new_utenti.py"
__author__ =    "luca"
__version__ =   "1.0.0"
__date__ =      "2011-12-23"

import hashlib
import re
from datetime           import datetime
from sqlalchemy.orm     import object_session, column_property
from sqlalchemy.orm.util import has_identity
from waonline.utils.utility import random_password
from waonline.core.ete_dumbobj import EteDumbObj
from waonline.core.users.decorators  import userscombomethod
from waonline.core.tabelle import Base, Column, Integer, ForeignKey, \
        String, Boolean, Table, DateTime, relationship, PickleType

from waonline.core import waologgin as wl


class StudioCodeError(BaseException): pass
class UserLevelError(BaseException): pass

def se(username):
    return username.replace("'", r"\'")


ERR_LIST = ['USER_DONT_EXISTS', 'USER_EXISTS', 'USER_CREATED',
'USER_PARENT_ERROR', 'USER_LEVEL_ERROR', 'USER_NO_LEVEL',
'USER_ADMIN_LEVEL_ERROR']

class DittaResult(object):
    def __init__(self, code, obj=None):
        self.code = code
        self.obj = obj

    def __eq__(self, other):
        return self.code == other

    def __enter__(self):
        return self.obj

    def __exit__(self, exception_type, exception_val, trace):
       return True

    def __str__(self):
        return "%s:%s" % (self.code, ERR_LIST[self.code])

    def __repr__(self):
        if self.obj:
            return repr(self.obj)
        else:
            return self.__str__()

    def __call__(self):
        return self.obj

_utenti = Table('_new_utenti', Base.metadata,
                Column('utente_id', Integer, ForeignKey('new_utenti.id',
                    ondelete='CASCADE')),
                Column('genitore_id', Integer, ForeignKey('new_utenti.id')),
                #useexisting=True
                )

class NewUtente(Base):
    """ Classe per la tabella utenti """
    __tablename__ = 'new_utenti'
    #__table_args__ = {'useexisting': True}
    __useexisting__ = True
    #: Costante per il livello ADMIN
    ADMIN = 100
    #: Costante per il livello SERVICE
    SERVICE = 80
    #: Costante per il livello USTUDIO (NewUtente Studio)
    USTUDIO = 60
    #: Costante per il livello UAZIENDA (NewUtente Azienda)
    UAZIENDA = 40
    #: Costante per il livello AZIENDA
    AZIENDA = 0
    #: Costanti per ritorno `NewUtente.create_user`
    USER_DONT_EXISTS = 0
    USER_EXISTS = 1
    USER_CREATED = 2
    USER_PARENT_ERROR = 3
    USER_LEVEL_ERROR = 4
    USER_NO_LEVEL = 5
    USER_ADMIN_LEVEL_ERROR = 6

    #: Id dell'utente.
    id = Column(Integer, primary_key=True, autoincrement=True)
    #: nome utente per l'account.
    username = Column(String(50), unique=True)
    #: password
    _password = Column(String(100), default="******")
    #: Id dell'utente nel database di riferimento
    db_id = Column(Integer, default=0)
    #: Nome del database di riferimento
    db_ref = Column(String(25), default='WA')
    #: Id dell'utente nello schema di riferimento
    schema_id = Column(Integer, default=0)
    #: Nome dello schema di riferimento
    schema_ref = Column(String(25), default='public')
    #: Id dell'utente studio
    studio_id = Column(Integer, default=0)
    #: Codice dello studio
    studio_codice = Column(String(3))
    #: data di scadenza password (intero YYYYMMDD).
    scad_password = Column(Integer)
    #: mesi di validità della password.
    mesi_password = Column(Integer)
    #: livello dell'utente (:data:`AZIENDA`, :data:`UAZIENDA` ... etc).
    livello = Column(Integer)
    #: codice dell utente per :class:`UAzienda` non e rilevante  mentre
    #: e imoprtante nel caso di :class:`Azienda`, :class:`UStudio` e
    #: :class:`Service`.
    codice = Column(String(10))
    #: Descrizione dell'utente.
    descrizione = Column(String(100))
    #: Autorizzazione sulle anagrafiche (1 carattere 'A','C','' | 'N')
    auto_anagrafiche = Column(String(1), default='A')
    #: Autorizzazione sulle presenze (1 carattere 'A','C','' | 'N')
    auto_presenze = Column(String(1), default='A')
    #: Autorizzazione dell'utente a inviare anagrafiche (True | False)
    is_sped_anag = Column(Boolean, default=True)
    #: Autorizzazione dell'utente a inviare presenze (True | False)
    is_sped_pres = Column(Boolean, default=True)
    #: Autorizzazione dell'utente a ricevere file (True | False)
    is_ricezione = Column(Boolean, default=True)
    is_segreteria = Column(Boolean, default=False)
    # Preferenze
    #: Seleziona tutti i dipendenti nella scermata di selezione del calendario
    #: (True | False)
    seleziona_tutti = Column(Boolean, default=None)
    #: Imposta il calendario verticale come predefinito per il singolo
    #: dipendente
    calendario_verticale = Column(Boolean, default=None)
    #: Ordinamento per la lista dei dipendenti
    ordinamento = Column(Integer, default=1)
    #: Indica se il tipo di ordinamento dovrà essere raggruppato o meno
    #: (True | False)
    tipo_ordinamento = Column(Boolean, default=None)
    #: Indica se le ore devono essere espresse in sessantesimi (True | False)
    is_sessantesimi = Column(Boolean, default=None)
    #: Data di creazione dell'utente
    data_creazione = Column(DateTime, default=datetime.now)
    #: Data di modifice dell'utente
    data_modifica = Column(DateTime, onupdate=datetime.now)
    #: .. warning:: Non lo so chiedere ad Oscar
    utente_modifica = Column(String(50))
    # Relazioni many-to-many
    #: Lista di oggetti :class:`NewUtente` genitori dell'utente, in genere solo
    #: gli utenti :class:`Azienda` possono avere più di un genitore.
    genitori_rel = relationship('NewUtente', secondary=_utenti,
                                primaryjoin=_utenti.c.utente_id==id,
                                secondaryjoin=_utenti.c.genitore_id==id,
                                backref='figli_rel')# lazy='joined')
    _genitori_rel = None
    _id_genitori_rel = None
    _figli_rel = None
    _id_figli_rel = None

    # Relazioni one-to-many
    #: lista di oggetti :class:`NewUtentePref` contenenti le preferenze
    #: dell' utente (sostanzialmente le varabili d'ambiente di Webazienda)
    preferenze_rel = relationship('NewUtentePref' )#, lazy='joined')

    #: Id del genitore effettivo, Viene usato sopratutto dalla proprità
    #: preferenze per gli utenti di tipo :class:`Azienda`
    genitore_id = column_property(Column("genitore_id", Integer, default=0))

    __mapper_args__ = {'polymorphic_on': livello}


    #   Proprieta NON correlate al database
    stato = False
    #: Riferimento all'oggetto :class:`UStudio`.
    studio = None
    _ditte = None
    _codici_ditte = None
    _id_ditte = None
    _id_children = None
    _children = None
    _sessiondb = None

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def add_pref(self, name, value, genitore_id=None):
        """ aggiunge una preferenza, se esiste gia mette il segno piu per
            creare una lista
        """
        genitore_id = genitore_id or self.genitore_id or self.id
        add_plus = False
        lista = [p for p in self.preferenze_rel if p.genitore_id==genitore_id]
        for pr in lista:
            nname = pr.nome[pr.nome.find('+') + 1:]
            if nname == name and pr.valore == value:
                return
        for pr in lista:
            nname = pr.nome[pr.nome.find('+') + 1:]
            if nname == name:
                add_plus = True
                if pr.nome[0] != '+':
                    pr.nome = '+' + pr.nome
        if add_plus:
            name = '+' + name
        pref = NewUtentePref(name, value)
        pref.utente_id = self.id
        pref.genitore_id = genitore_id
        self.preferenze_rel.append(pref)
        self.sessiondb.add(pref)
        self.commit()


    def create_ditta(self, codice, genitore_id=None, **kwa):
        if genitore_id:
            genitore = self.get_by_id(genitore_id)
            if not genitore:
                return DittaResult(self.USER_PARENT_ERROR)
        else:
            genitore = self
        if genitore.livello != self.UAZIENDA:
            return DittaResult(self.USER_LEVEL_ERROR)
        codice = codice.zfill(7).upper()[:7]
        username = ("%s_%s" % (genitore.studio_codice, codice)).upper()
        ditta = self.get_by_username(username)
        if ditta:
            return DittaResult(self.USER_EXISTS, ditta)
        kwa['username'] = username
        kwa['password'] = random_password()
        kwa['studio_id'] = genitore.studio_id
        kwa['studio_codice'] = genitore.studio_codice
        kwa['db_ref'] = genitore.db_ref
        kwa['schema_ref'] = genitore.schema_ref
        kwa['codice'] = codice
        new_ditta = NewAzienda(**kwa)
        self.sessiondb.add(new_ditta)
        genitore.figli_rel.append(new_ditta)
        self.sessiondb.flush()
        new_ditta.db_id = new_ditta.db_id or new_ditta.id
        new_ditta.schema_id = new_ditta.schema_id or new_ditta.id
        self.sessiondb.commit()
        return DittaResult(self.USER_CREATED, new_ditta)

    def delete_user(self, user):
        """ Elimina l'utente """
        if isinstance(user, list):
            self.sessiondb.delete(NewUtente.id.in_(user))
        else:
            self.sessiondb.delete(user)
        self.sessiondb.commit()

    def update(self, **kwa):
        """ Aggiorna l'utente """
        #: recuperiamo le colonne e togliamo quelle che non sono normalmente
        #: modificabili.
        keys = self.__table__.columns.keys()
        map(keys.remove, ['username', '_password', 'db_id', 'db_ref',
            'schema_id', 'schema_ref', 'studio_id', 'studio_codice'])
        if kwa.get('livello') != None:
            kwa['livello'] = int(kwa['livello'])
        for k, v in kwa.items():
            if k in keys:
                setattr(self, k, v)
        genitore = self.get_genitore()
        if genitore:
            if self.livello >= genitore.livello:
                raise UserLevelError("Il livello deve essere inferiore "\
                        "a quello del genitore.")
        self.sessiondb.commit()

    def child_by_id(self, chid_id):
        """ ritorna un figlio dall'id """
        try:
            return [c for c in self.children if c.id == chid_id][0]
        except IndexError:
            return None

    @property
    def session_status(self):
#: transient:
        if object_session(self) is None and not has_identity(self):
            return "transient"
#: pending:
        elif object_session(self) is not None and not has_identity(self):
            return "pending"
#: detached:
        elif object_session(self) is None and has_identity(self):
            return "detached"
#: persistent:
        elif object_session(self) is not None and has_identity(self):
            return "persistent"

    @property
    def sessiondb(self):
        if self._sessiondb == None:
            wl.warn("SESSION STATUS: %s " % self.session_status)
            self._sessiondb = object_session(self)
            if self._sessiondb == None:
                wl.warn("SESSION PANIC!!!!")
                raise Exception("Session PANIC")
        return self._sessiondb

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.username)

    @property
    def password(self):
        """ (*getter*) Ritorna la password

        (*setter*)Imposta la password trasformandola in upper case """
        return self._password

    @password.setter
    def password(self, passwd):
        """ Imposta la password trasformandola in upper case """
        self._password = passwd

    def check_password(self, passwd):
        """ Verifica che la password corrisponda a quella salvata nel database.
        """
        #: Verifica la password 'sha' per compatibilita con i vecchi utenti
        if self._password == hashlib.sha1(passwd.upper()).hexdigest():
            try:
                #: Prova a reimpostare la password in chiaro
                self._password = passwd.upper()
                self.sessiondb.commit()
            except:
                pass
            return True
        else:
            return self._password.upper() == passwd.upper()

    @property
    def id_children(self):
        """ lista degli id dei figli """
        if self._id_children == None:
            self._id_children = [c.id for c in self.children]
        return self._id_children

    @property
    def children(self, clist=[]):
        """ lista dei figli """
        if self._children == None:
            self._children = self.sessiondb.query(NewUtente).\
                from_statement("select * from children_of(%d)" % self.id).all()
        return self._children

    @property
    def preferenze(self, genitore_id = None):
        """ Riferimento ad una lista contenente tutti gli oggetti
        :class:`NewUtentePref` dell'utente relativi al genitore
        (in caso di :class:`Azienda`) o relativi a se stesso (negli altri casi)
        """
        if self.livello == self.AZIENDA:
            genitore_id = genitore_id or self.genitore_id or self.id
        else:
            genitore_id = self.id
        return [p for p in self.preferenze_rel if p.genitore_id == genitore_id]

    def get_all_prefs(self, toLista=False):
        """ Ritorna tutte le preferenze partendo da quelle del :class:`NewUtente`
        di livello più alto gerarchicamente. In caso di preferenze con lo stesso
        nome verranno sostituite con quelle dell' :class:`NewUtente` del livello
        più basso.

        In caso di **filtri** (*FILTRO_NOME_FILTRO* ,
        *FILTRO_NO_NOME_FILTRO*) i valori verranno messi in una lista.
        Un eventuale *FILTRO_NO_NOME_FILTRO* al livello più basso andrà
        ad invalidare il relativo *FILTRO_NOME_FILTRO* con il medesimo valore
        del livello più alto e viceversa.

        Ritorna un oggetto :class:`wadecohelpers.EteDumbObj` con le preferenze
        oppure una lista di tuplet (*id*, *nome*, *valore*) se il parametro
        *toLista* e True.
        """
        pat = re.compile(r'^FILTRO_(?P<tipo>NO_){0,1}(?P<nome>\w+)')
        prefs = {}
        prefs_id = {}
        prefs_lv = {}
        count_ante = 0
        for utente in self.get_antenati() + [self]:
            prefs_idx = {}
            lpref = None
            for pref in utente.preferenze:
                if not pref.nome:
                    continue
                m = pat.match(pref.nome.upper())
                if m:
                    d = m.groupdict()
                    if not d['nome'] in prefs:
                        prefs[d['nome']] = {'si': [[]], 'no': [[]]}
                        prefs_id[d['nome']] = pref.id
                    ntipo = 'no' if d['tipo'] == 'NO_' else 'si'
                    try:
                        lpref = prefs[d['nome']][ntipo][count_ante]
                    except:
                        #prefs[d['nome']]['no'][count_ante] = []
                        prefs[d['nome']][ntipo].append([])
                        lpref = prefs[d['nome']][ntipo][count_ante]
                    if not pref.valore in lpref:
                        lpref.append(pref.valore)
                else:
                    nnome = pref.nome
                    if nnome.startswith('+'):
                        isLista = True
                        nnome = nnome[1:]
                    elif nnome.startswith('-'):
                        nnome = nnome[1:]
                        if nnome in prefs:
                            if isinstance(prefs[nnome], list):
                                if pref.valore in prefs[nnome]:
                                    prefs[nnome].remove(pref.valore)
                            else:
                                del prefs[nnome]
                        continue
                    else:
                        isLista = False
                    if not nnome in prefs:
                        if isLista:
                            prefs[nnome] = [pref.valore]
                        else:
                            prefs[nnome] = pref.valore
                        prefs_id[nnome] = pref.id
                        prefs_lv[nnome] = utente.livello
                        prefs_idx[nnome] = 0
                    else:
                        if not nnome in prefs_idx:
                            prefs_idx[nnome] = 0
                        if isLista:
                            if not isinstance(prefs[nnome], list):
                                prefs[nnome] = [prefs[nnome]]
                            if not pref.valore in prefs[nnome]:
                                prefs[nnome].append(pref.valore)
                        else:
                            if isinstance(prefs[nnome], list):
                                prefs[nnome][prefs_idx[nnome]] = pref.valore
                            else:
                                prefs[nnome] = pref.valore
                        prefs_idx[nnome] += 1
            if lpref:
                count_ante += 1
        if toLista:
            return zip(prefs_id.values(), prefs.keys(), prefs.values())
        else:
            return EteDumbObj(prefs)

    @property
    def impostazioni(self):
        return EteDumbObj({
            'seleziona_tutti': self.seleziona_tutti,
            'calendario_verticale': self.calendario_verticale,
            'orientamento': 'verticale' if self.calendario_verticale \
                    else 'orizzontale',
            'ordinamento': self.ordinamento,
            'tipo_ordinamento': self.tipo_ordinamento,
            'is_sessantesimi': self.is_sessantesimi,
            'is_segreteria': self.is_segreteria,
            'auto_anagrafiche': self.auto_anagrafiche,
            'auto_presenze': self.auto_presenze
        })

    @property
    def info(self):
        try:
            genitore_id = self.genitore_id or self.get_genitore().id
        except AttributeError:
            genitore_id = 1
        return EteDumbObj({
            'id': self.id,
            'codice': self.codice,
            'username': self.username,
            'descrizione': self.descrizione,
            'livello': self.livello,
            'genitore_id': genitore_id
        })

    def get_all_impostazioni(self):
        """ Ritorna un dizionario con le impostazioni per il client partendo da
        quelle dell' :class:`NewUtente` di livello più alto gerarchicamente.
        Sostituendole di volta in volta con quelle di livello più basso se
        presenti.
        """
        impostazioni = self.impostazioni
        for utente in self.get_antenati() + [self]:
            for chiave in impostazioni.keys():
                if hasattr(utente, chiave):
                    valore = getattr(utente, chiave)
                    if not valore == None:
                        impostazioni[chiave] = valore
        impostazioni.orientamento = 'verticale' if impostazioni.calendario_verticale \
                else 'orizzontale'
        return impostazioni

    def ditta_by_codice(self, codice):
        try:
            return [d for d in self.ditte if d.codice==codice][0]
        except IndexError:
            return None

    def ditta_by_id(self, id):
        try:
            return [d for d in self.ditte if d.id==id][0]
        except IndexError:
            return None

    @property
    def codici_ditte(self):
        """ ritorna i codici delle ditte """
        if self._codici_ditte == None:
            self._codici_ditte = [c.codice for c in self.ditte]
        return self._codici_ditte

    @property
    def id_ditte(self):
        """ ritorna gli id delle ditte """
        if self._id_ditte == None:
            self._id_ditte = [c.id for c in self.ditte]
        return self._id_ditte

    @property
    def ditte(self):
        """ Ritorna una lista con tutte le ditte (:class:`Azienda`) per
        l'utente, comprese quelle dei relativi figli.
        """
        if self._ditte == None:
            self._ditte = [a for a in self.children\
                    if a.livello == self.AZIENDA]
        return self._ditte

    def has_ditta(self, codici):
        """ Ritorna True se il codice della ditta appartiene ad una delle ditte
        dell'utente, altrimenti ritorna il primo codice non valido.
        *codici* può essere sia una stringa che una lista di stringhe.
        """
        if not isinstance(codici, list):
            codici = [codici]
        for codice in codici:
            if not codice.upper() in [d.codice.upper() for d in self.ditte]:
                return codice.upper()
        return True

    @property
    def id_genitori_rel(self):
        """ Lista degli id genitori """
        if self._id_genitori_rel == None:
            self._id_genitori_rel = [g.id for g in self.genitori_rel]
        return self._id_genitori_rel

    @property
    def id_figli_rel(self):
        """ Lista degli id figli """
        if self._id_figli_rel == None:
            self._id_figli_rel = [f.id for f in self._figli_rel]
        return self._id_figli_rel

    def get_genitore(self, genitore_id = None):
        """ Ritorna l':class:`NewUtente` genitore attuale o None se non ha genitori
        """
        if not self.genitori_rel:
            return None
        genitore_id = genitore_id or self.genitore_id
        if genitore_id:
            genitore = [g for g in self.genitori_rel if g.id == genitore_id]
            if genitore:
                return genitore[0]
            else:
                return None
        else:
            return self.genitori_rel[0]

    def get_antenati(self):
        """ Ritorna una lista con tutti gli utenti a partire dal livello più
        alto gerarchicamente.
        """
        lista = []
        ant = self.get_genitore()
        if ant:
            lista.append(ant)
            while ant.get_genitore():
                ant = ant.get_genitore()
                if ant:
                    lista.append(ant)
                else:
                    break
        lista.reverse()
        return lista


    def _q(self, query, *qfilters, **kwargs ):
        """ Wrapper per self.sessiondb.query con supporto filtri multiplo che
        vuol dire che la query **_query(Ditta).filter(Ditta.codice==codice)**
        può essere più comodamente scritta **_query(Ditta,
        Ditta.codice==codice)**
        """
        if query.__class__.__name__ == 'InstrumentedAttribute':
            query = [query]
            query.extend(qfilters)
            query = tuple(query)
            q = self.sessiondb.query(*query, **kwargs)
        else:
            q = self.sessiondb.query(query, **kwargs)
            for qfilter in qfilters:
                q = q.filter(qfilter)
        return q

    def _ex(self, query, commit=False):
        """ Wrapper for self.sessiondb.execute """
        q = self.sessiondb.execute(query)
        if commit:
            self.sessiondb.commit()
        return q

    def commit(self):
        self.sessiondb.commit()

    #: combomethods
    @userscombomethod
    def create_user(self, uname, passwd, genitore_id=None, sessiondb=None, **kwa):
        """ Crea un utente """
        if self.user_exists(uname, sessiondb=sessiondb):
            return self.USER_EXISTS
        if genitore_id:
            genitore = self.get_by_id(genitore_id, sessiondb=sessiondb)
            if not genitore:
                return self.USER_PARENT_ERROR
        elif isinstance(self, NewUtente):
            genitore = self
        else:
            genitore = None
        livello = int(kwa.get('livello', self.AZIENDA))
        kwa['livello'] = livello
        #if level < self.ADMIN and genitore == None:
            #return self.USER_ADMIN_LEVEL_ERROR
        if genitore and livello >= genitore.livello:
            return self.USER_LEVEL_ERROR
        cls = CLASSI.get(livello)
        if not cls:
            return self.USER_NO_LEVEL
        if livello == self.USTUDIO:
            try:
                kwa['studio_codice'] = kwa['codice']
                kwa['auto_presenze'] = None
                kwa['auto_anagrafiche'] = None
            except KeyError:
                raise StudioCodeError("Un utente Studio deve avere un codice")
                return
            kwa['db_ref'] = kwa.get('db_ref', "WA_%s" % kwa['studio_codice'])
            kwa['schema_ref'] = kwa.get('schema_ref', "studio_%s" % \
                    kwa['studio_codice'])
        elif genitore:
            kwa['studio_id'] = genitore.studio_id
            kwa['studio_codice'] = genitore.studio_codice
            kwa['db_ref'] = kwa.get('db_ref', "WA_%s" % genitore.studio_codice)
            kwa['schema_ref'] = kwa.get('schema_ref', "studio_%s" % \
                genitore.studio_codice)
        kwa['password'] = passwd
        kwa['username'] = uname
        new_user = cls(**kwa)
        sessiondb.add(new_user)
        if genitore:
            genitore.figli_rel.append(new_user)
        sessiondb.flush()
        new_user.db_id = new_user.db_id or new_user.id
        new_user.schema_id = new_user.schema_id or new_user.id
        if livello == self.USTUDIO:
            new_user.studio_id = new_user.id
        sessiondb.commit()
        return new_user

    @userscombomethod
    def get_orphans(self, sessiondb=None, max=60, min=0):
        return sessiondb.query(NewUtente).from_statement(
                "select * from users_orphans(%d, %d)" % (max, min)).all()

    @userscombomethod
    def user_exists(self, username, sessiondb=None):
        return sessiondb.execute("SELECT user_exists('%s');" % se(username)).scalar()

    @userscombomethod
    def login(self, username, password, sessiondb=None):
        """ Tenta di autentificare l'utente con *username* e *password*. Ritorna
        l'utente se l'autentificazione va a buon fine altrimenti ritorna None.
        """
        user = sessiondb.query(NewUtente).filter(
                NewUtente.username==username.upper()).filter(
                        NewUtente.livello > NewUtente.AZIENDA)
        print str(user)
        user = user.first()
        print user
        if user:
            print user.password
            print password
            print user.check_password(password.upper())
            if user.check_password(password.upper()):
                return user
        return None

    @userscombomethod
    def get_by_id(self, uid, sessiondb=None):
        """ Ritorna l'utente recuperandolo dall'*id*."""
        #: in caso uid sia una lista ritorno la lista di utenti
        if isinstance(uid, list):
            return sessiondb.query(NewUtente).filter(
                    NewUtente.id.in_(uid)).all()
        return sessiondb.query(NewUtente).get(uid)

    @userscombomethod
    def get_by_username(self, username, sessiondb=None):
        """ Ritorna l'utente recuperandolo dall'*username*."""
        return sessiondb.query(NewUtente).filter(
               NewUtente.username==username.upper()).first()

    @userscombomethod
    def get_by_property(self, property_name, value, sessiondb=None):
        """ Ritorna l'utente confrontando la proprieta *property_name* con
            il valore *value*.
        """
        return sessiondb.query(NewUtente).filter(
               getattr(NewUtente, property_name)==value).first()


class NewUtentePref(Base):
    """ Classe per le preferenze dell':class:`NewUtente`
    """
    __tablename__ = 'new_utenti_pref'
    #__table_args__ = {'useexisting': True}
    #useexisting = True
    #: id della preferenza nel database
    id = Column(Integer, primary_key=True, autoincrement=True)
    #: id dell':class:`NewUtente` di appartenenza
    utente_id = Column(Integer, ForeignKey('new_utenti.id',
        ondelete='CASCADE'))
    #: id del genitore relativo all':class:`NewUtente`
    genitore_id = Column(Integer)
    #: Nome della preferenza
    nome = Column(String(50))
    #: Valore della prefenza
    valore = Column(PickleType)

    def __init__(self, nome, valore, genitore_id = 0):
        self.nome = nome
        self.valore = valore
        self.genitore_id = genitore_id

    def __repr__(self):
        return "%s: %s" % (self.nome, self.valore)


class NewAdmin(NewUtente):
    """ Classe polimorfata per la gestione :class:`NewUtente` di lievello
    :data:`NewUtente.ADMIN`
    """
    __mapper_args__ = {'polymorphic_identity': NewUtente.ADMIN}


class NewService(NewUtente):
    """ Classe polimorfata per la gestione :class:`NewUtente` di lievello
    :data:`NewUtente.SERVICE`
    """
    __mapper_args__ = {'polymorphic_identity': NewUtente.SERVICE}


class NewUStudio(NewUtente):
    """ Classe polimorfata per la gestione :class:`NewUtente` di lievello
    :data:`NewUtente.USTUDIO`
    """
    __mapper_args__ = {'polymorphic_identity': NewUtente.USTUDIO}
    @property
    def studio(self):
        """ Nel caso di :class:`UStudio` fa riferimento a se stesso """
        return self

class NewUAzienda(NewUtente):
    """ Classe polimorfata per la gestione :class:`NewUtente` di lievello
    :data:`NewUtente.UAZIENDA`
    """
    __mapper_args__ = {'polymorphic_identity': NewUtente.UAZIENDA}
    @property
    def studio(self):
        """ Ritorna un riferimento alloggetto :class:`UStudio` relativo;
        nonostante :class:`UAzienda` debba avere un solo genitore la proprietà
        ritorna la lista in caso di genitori multipli.
        """
        lista = []
        for genitore in self.genitori_rel:
            if not genitore in lista:
                lista.append(genitore)
        if len(lista) == 1:
            return lista[0]
        return lista            # 

class NewAzienda(NewUtente):
    """ Classe polimorfata per la gestione :class:`NewUtente` di lievello
    :data:`NewUtente.AZIENDA`
    """
    __mapper_args__ = {'polymorphic_identity': NewUtente.AZIENDA}
    @property
    def studio(self):
        """ Ritorna un riferimento alloggetto :class:`UStudio` relativo al
        proprio genitore attivo. :data:`NewUtente.genitore_id` deve essere diverso
        da 0, e dovrebbe esserlo visto che normalmente di accede a
        :class:`Azienda` tramite :class:`UAzienda`. Ad ogni modo in caso
        contrario tenta di recuperare lo :class:`UStudio` studio per tutti i
        genitori e ritorna una lista.
        """
        if self.genitore_id:
            return self.get_genitore()
        else:
            lista = []
            for genitore_id in self.id_genitori_rel:
                genitore_db = self.sessiondb.query(NewUtente).filter(NewUtente.id == genitore_id).first()
                if not genitore_db.studio in lista:
                    lista.append(genitore_db.studio)
            if len(lista) == 1:
                return lista[0]
            return lista

#: Dizionario con le classi in relazione al livello
CLASSI = {0: NewAzienda, 40: NewUAzienda, 60: NewUStudio, 80: NewService,
        100: NewAdmin}

def get_class_from_level(livello):
    """ Recupera il costruttore della classe per l'utente in base al livello
    """
    try:
        return CLASSI[livello]
    except KeyError:
        return None
