# coding: latin-1
"""
====================================
Helpers per i decoratori di Waonline
====================================

:File: wadecohelpers.py
:Author: Luca Zaccaria luca.z@essepaghe.it
:Date: 10/11/2010

Modulo Helper per :class:`wadecorators.WaAuth`.

In questo modulo vanno definite le funzioni di *helper* per il decoratore
:class:`wadecorators.WaAuth`

Ongi funzione deve avere i due parametri **self** e **args** che sono
rispettivamente il riferimento all'istanza del decoratore :class:`WaAuth`
chiamante e il dizionario di attributi per l'argomento in analisi.

Tenendo persente che in **self.base**  disponibile la classe chiamante per la
la verifica del metodo; quello che normalmente  usato come **self** nella
funzione stessa.  Quindi, per esempio, il codice ditta che all'interno della
funzione  **self._cod_ditta** sar disponibile con nella *funzione helper* come
**self.base._cod_ditta**\.

**args**  sempre un dizionario, anche se vuoto e pu essere modificato per
forzare il comportamento di :class:`wadecorators.WaAuth` .

.. note::  necessario ritornare :data:`wadecorators.WaAuth.non_valido`
    per indicare a :class:`wadecorators.WaAuth` che ci sono stati problemi
    e si comporti di conseguenza.

**Esempi**
    Controlla che il nome utente sia mario, in caso contrario forza l'errore
    e anche il nome.

    Definiamo il nostro helper nel modulo :mod:`wadecohelpers` ::

        def nome(self,args):
            # Accediamo al primo dipendente nella lista e verifichiamo che
            # si chiami mario
            if self.base._dipti[0].nome.lower() != "mario":
                # se il dipendente non si chiama mario forziamo l'errore
                args['err'] = "Il nome deve essere mario"

                # e forziamo il valore per il parametro di ritorno
                argf['forza'] = 'mario'

                # alla fine ritorniamo non_valido per dire al decoratore che ci
                # son stati problemi
                return self.non_valido
            else:
                # Nel caso il dipendente di chiami mario torniamo il nome
                return self.base._dipti[0].nome

    Creiamo ora la nostra classe con la nostra funzione che utilizzi
    :class:`wadecorators.WaAuth` insieme al nuovo helper. ::

        from waonline.gestione.rbase         import RBase
        from waonline.core.wadecorators  import WaAuth


        class GestMario(RBase):

            @WaAuth()
            def solo_mario(self, nome):
                self._addMess("Il dipendente si chiama %s" % nome)

        # ...


    Nell'esempio qui sopra abbiamo forzato l'errore *err* con una stringa quindi
    :class:`wadecorators.WaAuth` utilizzer quello come messaggio d'errore da
    visualizzare. Abbiamo anche forzato il valore con *forza* quindi
    :class:`wadecorators.WaAuth` segnalera l'errore (in quanto forzato) ma
    riterr comunque valido il metodo ritornando il valore "mario"
    per l'argomento *nome*.


"""

from waonline.utils.utility import formatData
from waonline.core.ete_dumbobj import EteDumbObj
from waonline.utils.objutils   import fromJSON

def json_to_dumbobj(valore):
    """Ritorna un oggetto ``EteDumbObj`` convertendolo da json"""
    return EteDumbObj(fromJSON(valore))

def to_list(valore):
    """ Ritorna il valore come lista, in caso che non lo sia, ovviamente.
    Viene utilizzato come 'tipo'
    """
    if isinstance(valore, list):
        return valore
    else:
        return [valore]

def to_upper(valore):
    return str(valore).upper()

def dbl_date(valore):
    """ Fa la doppia conversione di una data per essere sicuro del formato """
    return formatData(formatData(valore),'int')

def get_get(self, vargs):
    """ Ritorna il nome della varaibile se  presente negli argomenti oppure
    None basandosi su vargs['get']. in caso vargs['get'] sia una lista ritorna
    il primo nome valido.
    """
    get = vargs.get('get')
    if not get:
        return None
    elif isinstance(get, list):
        for var in get:
            if var in self.base.args:
                return var
    else:
        if get in self.base.args:
            return get
    return None

def _dipti_id(self, args):
    """ Verifica la presenza e i diritti per i dipendenti """
    args['err'] = args.get('err', True)  # forziamo gli errori
    if self.base._dipti_id:
        if (self.base.user != None):
            if not self.base._checkDipe(self.base._dipti_id):
                self.add_err("Non si dispongono delle autorizzazioni per uno o\
                        piu' dipendenti", vargs=args)
                return self.non_valido
            else:
                return self.base._dipti_id
        else:
            self.add_err("Non si dispongono delle autorizzazioni per \
                    l'utente", vargs=args)
            return self.non_valido
    else:
        self.add_err("E' richiesto almeno un dipendente", vargs=args)
        return self.non_valido


def cod_ditta(self, args):
    """ Verifica la presenza e i diritti per il codice ditta """
    args['err'] = args.get('err', True)  # forziamo gli errori
    self.base._cod_ditta = self.base.ga(get_get(self, args) or 'cod_ditta')
    if self.base._cod_ditta:
        self.base._cod_ditta = self.base._cod_ditta.zfill(7)
        if (self.base.user != None):
            verifica = self.base.user.has_ditta(self.base._cod_ditta)
            if verifica != True:
                self.add_err("Non si dispongono di autorizzazioni per la \
                        ditta %s" % verifica, vargs=args)
                return self.non_valido
        else:
            self.add_err("Non si dispongono delle autorizzazioni per l'utente"\
                    , vargs=args)
            return self.non_valido
    else:
        self.add_err(args, "E' richiesta una ditta di riferimento")
        return self.non_valido
    return self.base._cod_ditta

def periodo(self, args):
    """ Verifica se c' il periodo """
    args['err'] = "Manca il periodo"  # forziamo gli errori
    return self.base._periodo or self.non_valido

