# coding: latin-1
"""
=======================
Decoratori per waonline
=======================

:File: wadecorators.py
:Author: Luca Zaccaria luca.z@essepaghe.it
:Date: 10/11/2010
"""

import inspect
from functools import wraps

from waonline.core import wadecohelpers
from waonline.core.ete_dumbobj import EteDumbObj


class  WaAuth(object):
    """ Questo modulo serve a decorare le funzioni di classi derivate da RBase in
    webazienda online in maniera tale da definire, tramite gli argomenti della
    funzione stessa o tramite gli argomenti del decoratore, quali sono le variabili
    passate da wademon necessarie e renderle disponibili come per una normale
    funzione.

    .. data:: DITTA

        Costante per la verifica del codice ditta e della sua validit per
        l'utente attuale. Da usare come argomento del decoratore.

    .. data:: DIPTE

        Costante per la verifica degli **id** dei dipendenti e della loro
        validit per l'utente attuale. Da usare come argomento del decoratore.

    .. data:: PERIODO

        Costante per la verifica del periodo. Da usare come argomento del
        decoratore.

    .. data:: INT

        Costante per il recupero di un numero intero con forzatura dell'errore.
        Da usare come valore di default per un argomento del metodo

    .. data:: INT0

        Costante per il recupero di un numero intero, senza forzatura
        dell'errore con valore di default a 0

    .. data:: non_valido

        Costante per indicare il valore non valido uscendo da una verifica o
        da una funzione di verifica (vedi :mod:`wadecohelpers`)

    La classe :class:`WaAuth` gestisce essenzialmente 3 tipi di argomenti:

        - della funzione
        - della funzione con valori di default
        - del decoratore

    **Argomenti della funzione**
       Vengono trattati esenzialmente come gli argomenti di una funzione.
       :class:`WaAuth` ne verifica la presenza (con :func:`rbase.RBase.getArg`)e
       in caso di valore mancante di uno degli argomenti restituisce *False* e
       **NON** esegue la funzione. Se la verifica di :class:`WaAuth` va a buon
       fine i valori saranno presenti per i rispettivi argomenti.

    **Argomenti della funzione con valori di default**
        Vengono trattati esenzialmente come gli argomenti di una funzione con
        valori di default quindi in caso di valore mancante viene considerato il
        valore di default. E' possibile specificare un dizionario come valore di
        default per definire il comportamento di :class:`WaAuth`.

    **Argomenti del decoratore**
        Gli argomenti del decoratore vengono passati come una lista di stringhe
        o di dizionari e valori sono resi disponibili nell'argomento "result"
        della funzione, se presente, altrimenti nell'oggetto self.result.
        Non fanno controlli particolari, se non specificato, :class:`WaAuth` si
        limita a renderli disponibili se li trova (None se il valore non 
        presente).
        Sostanzialmente viene utilizzato per impostare controlli standard come
        :data:`DITTA` oppure :data:`DIPTE` che non necessariamente necessitano
        di un valore visto che  gi disponibile in
        :data:`rbase.RBase._cod_ditta` e in :data:`rbase.RBase._dipti_id`.

    **Attributi dei dizionari**
        E' possibile specificare per ogni argomento un dizionario con alcuni
        attributi per controllare il comportamento di :class:`WaAuth`:

        +--------------+--------------------------------------------------------+
        | Attributo    | Valori                                                 |
        +==============+========================================================+
        |``'err'``     | [True | str] Forza la segnalazione di un errore        |
        |              | recuperando i valori per gli argomenti Impostato a True|
        |              | visualizzer un messaggio di errore standard mentre se |
        |              | err  una stringa verr utilizzata come messaggio      |
        |              | d'errore.                                              |
        +--------------+--------------------------------------------------------+
        |``'tipo'``    | [type | funzione] :class::`WaAuth` utilizzer il       |
        |              | contenuto di questo attributo per tentare una          |
        |              | conversione del dato.                                  |
        +--------------+--------------------------------------------------------+
        |``'default'`` | [any] In caso di mancanza di dato :class:`WaAuth`      |
        |              | utilizzer questo come valore.                         |
        +--------------+--------------------------------------------------------+
        |``'forza'``   | [True | any] In caso di dato mancante o di errore di   |
        |              | conversione :class:`WaAuth` utilizzer "default" se    |
        |              | forza  uguale a True oppure il valore contenuto in    |
        |              | forza.                                                 |
        +--------------+--------------------------------------------------------+
        |``'get'``     | [str | list] nome o lista di nomi dell'argomento       |
        |              | possibile da cui recuperare il valore in alternativa al|
        |              | nome del parametro, il valore in get, in caso ha       |
        |              | priorit sul valore del parametro.                     |
        +--------------+--------------------------------------------------------+

    **Funzioni di controllo tramite :mod:`wadecohelpers`**
         possibile scrivere delle apposite funzioni di controllo nel modulo
        wadecohelpers.py per fare controlli pi complessi sugli argomenti.
        (vedere wadecohelpers per i dettagli)

    **Esempi**

        Utilizza il pattern di defaul :data:`DITTA` per verificare la presenza
        del valore e i diritti sulla ditta; in caso positivo il valore sar
        disponibile in self.result.cod_ditta, altrimenti verr segnalato
        l'errore e non verr eseguita la funzione. ::

            @WaAuth(WaAuth.DITTTA)
            def mostra_codice_ditta(self):
                self.adMess(self.result.cod_ditta)

        Forza il controllo dei dipendenti (utilizza
        :func:`wadecohelpers.id_dipte`) e rende disponibile il valore per
        l'argomento id_dipte poi tenta di recuperare le ore_teoriche e in caso
        di mancato valore o di errore forza ore_teoriche a 0 e lo rende
        disponibile come argomento.::

            @WaAuth()
            def dipendenti_ore(self, id_dipte, ore_teoriche={'tipo':float, 'forza': 0}):
                for dipte in id_dipte:
                   self._addMess("id dipendente %s" % dipte)
                self._addMess("ore teoriche %02.2f" % ore_teoriche)

        Utilizza i pattern di default per controllare e recuperare ditta e
        dipendenti poi li mette a disposizione come propriet dell'oggetto result
        indicato come parametro (utilizza :class:`rbase.RBase._ditta` e
        :class:`rbase.RBase._dipti` per le altre cose).::

            @WaAuth(WaAuth.DITTTA + WaAuth.DIPTE)
            def showres(self, result):
                self._addMess("Ditta: %s - %s" % (self.result.cod_ditta, self._ditta.descrizione))
                self._addMess('id dipendenti %s ' % self.result.id_dipte)
                for dipe in self._dipti:
                    self._addMess('Didpendente %s' % dipe.nominativo)

        Un untilizzo pi esteso dell'attributo tipo
        In questo caso usa json.loads per convertire il dato del primo argomento
        argomentoin un dizionario python e poi una funzione lambda per convertire
        il dato del secondo argomento in un intero e moltiplicarlo per 3::

            @WaAuth()
            def jsoneno(self,
                dati=dict(tipo=json.loads),
                quanti={'tipo':lambda x: int(x)*3}):
                self._addMess(dati["dipendente"]["nome"])
                self._addMess("per tre: %s" % quanti)


    L'attributo tipo permette una manipolazione del dato un po' pi complessa
    della semplice conversione, ma il dato in arrivo deve essere presente o
    avere un valore di default. Non pu quindi sostituire completamente le
    funzioni di verifica implementabili nel modulo wadecohelpers.

    Attualmente se WaAuth appende l'oggetto result alla classe chiamante questo
    rimane inalterato fino alla prossima chiamta di WaAuth con argomenti del
    decoratore oppure, ovviamente, fino alla cancellazione o alla modifica manuale.
    """

    DITTA = [{"cod_ditta":{'err':True}}]
    DIPTE = [{"id_dipte":{'err':True, 'tipo':int}}]
    PERIODO = [{"periodo":{'err':'Manca il periodo', 'tipo':int}}]
    INT = {'err': True, 'tipo': int}
    INT0 = {'err': True, 'tipo': int, 'default': 0}
    STR = {'err': True, 'tipo': str}
    STRV = {'err': True, 'tipo': str, 'default': ''}
    UPPER = {'err': True, 'tipo': wadecohelpers.to_upper}
    UPPERV = {'err': True, 'tipo': wadecohelpers.to_upper, 'default': ''}
    JEDO = {'err': True, 'tipo': wadecohelpers.json_to_dumbobj}
    JEDOV = {'err': True, 'tipo': wadecohelpers.json_to_dumbobj, 'default': '{}'}

    non_valido = "wY3=16oA*Py8"

    def __init__(self, policy=None):
        self.policy = policy or []
        self.func = None
        self.base = None
        self.kwargs = None
        self.nvars = None
        self.args = None
        self._result = EteDumbObj()

    def __call__(self, func, *args, **kwargs):
        self.func = func
        @wraps(func)
        def wrapped_f(base, *args, **kwargs):
            self.base = base
            if self.check_auth(self.base, self.func, *args, **kwargs):
                self.args = tuple(self.args)
                return self.func(self.base, *self.args, **self.kwargs)
            else:
                return False
        return wrapped_f

    def check_auth(self, base, func, *args, **kwargs):
        """Verifica le autorizzazioni"""
        self.base = base
        self.nvars, self.args, self.kwargs, defaults = inspect.getargspec(func)
        self.args = list(args)
        if not self.kwargs:
            self.kwargs = {}
        if 'self' in self.nvars:
            self.nvars.remove('self')
        if len(self.args) < len(self.nvars):
            if defaults:
                defaults = defaults[-(len(self.nvars)-len(self.args)):]
                self.kwargs = dict(dict(zip(self.nvars[-len(defaults):],
                    defaults)))
                self.nvars = self.nvars[:-len(defaults)]
        self.kwargs.update(kwargs)
        lista = self.nvars[len(self.args):] + [{c:v} for c, v in \
                self.kwargs.items() if c not in self.nvars] + self.policy
        self.args.extend(self.nvars[len(self.args):])
        for policy in lista:
            if isinstance(policy, str):
                if policy == "result":
                    continue
                variabile = policy
                vargs = {}
            elif isinstance(policy, dict):
                if len(policy.keys()) == 1:
                    variabile = policy.keys()[0]
                    vargs = policy[variabile]
                    if not isinstance(vargs, dict):
                        vargs = {'default': vargs}
                    #variabile = vargs.get('get', variabile)
                else:
                    variabile = policy.get('get')
                    if not variabile:
                        raise Exception("Manca il nome per una variabile.\
                                Specificarlo direttamente o con l'attributo \
                                \"get\"")
                    vargs = policy

            if hasattr(wadecohelpers, variabile):
                valore = getattr(wadecohelpers, variabile)(self, vargs)
                vargs['from_method'] = valore
            else:
                valore = self.base.ga(wadecohelpers.get_get(self, vargs) or
                        variabile,
                        default=vargs.get('default', self.non_valido),
                        nolist=vargs.get('nolist', False),
                        forcelist=vargs.get('forcelist', False)
                )
                if valore == None and vargs.get('default') != None:
                    valore = vargs.get('default')
            if not self.set_var(variabile, valore, vargs):
                return False
        if len(self._result.keys()):
            if 'result' in self.nvars:
                self.args[self.nvars.index('result')] = self._result
            else:
                self.base.result = self._result
        return True

    def set_var(self, variabile, valore, vargs):
        """Imposta le variabili"""
        if variabile in self.nvars:
            if valore == self.non_valido:
                vargs['err'] = vargs.get('err', True)
                self.add_err("Errore per la variabile %(var)s", variabile,
                        vargs=vargs)
                return False
            self.args[self.nvars.index(variabile)] = valore
            return vargs.get('from_method', True)
        else:
            if vargs:
                if valore == self.non_valido:
                    self.add_err(None, variabile, valore, vargs)
                    forza = vargs.get('forza', self.non_valido)
                    if forza == True:
                        valore = vargs.get('default', None)
                    elif forza != self.non_valido:
                        valore = forza
                    else:
                        return False
                if 'tipo' in vargs:
                    try:
                        if isinstance(valore, list):
                            for i in range(len(valore)):
                                valore[i] = vargs.get('tipo', str)(valore[i])
                        else:
                            valore = vargs.get('tipo', str)(valore)
                    except ValueError, err:
                        self.add_err("%s " % err + \
                            " [%(var)s %(val)s]", variabile, valore, vargs)
                        return False
        if variabile in self.kwargs.keys():
            self.kwargs[variabile] = valore
        else:
            self._result[variabile] = valore
        return vargs.get("from_method", True)

    def add_err(self, errore=None, variabile=None, valore=None, vargs=None):
        """Aggiunge l'errore"""
        verr = {"val": str(valore).replace(self.non_valido, ''),
                "var": variabile}
        if not errore:
            if not variabile in self.base.args:
                errore = "Valore mancante per la variabile %(var)s" % verr
            else:
                errore = "Valore non valido per la variabile %(var)s \
(%(val)s)" % verr
        else:
            try:
                errore = errore % verr
            except Exception, err:
                errore = err
        if vargs:
            serrore = vargs.get('err')
            if serrore == True:
                self.base._addErr(errore)
            elif serrore:
                self.base._addErr(serrore % verr)


class combomethod(object):
    def __init__(self, method):
        self.method = method

    def __get__(self, obj=None, objtype=None):
        @wraps(self.method)
        def _wrapper(*args, **kwargs):
            if obj is not None:
                return self.method(obj, *args, **kwargs)
            else:
                return self.method(objtype, *args, **kwargs)
        return _wrapper
