#!/usr/bin/env python
# coding: latin-1
# File: webdaemon.py
# Author: Marco Dalla Stella
# Date: 07/05/2010
"""
Questo modulo fa partire il demone tornado. Opzioni utili sono:

 --demonize=0 (default 1). Questa opzione evita il fork del processo che
      quindi resta in foreground e permette il debug tramite il modulo
      ipdb se chiamato da una funzione: import ipdb; ipdb.set_trace()

Se si vuole invece debuggare con ipython direttamente db_manager o db_session
si pu fare partire cos:

   ipython wademon.py

e dal prompt si potr interagire direttamente

"""
#### Imports & defaults
import daemon
import locale
import os
import sys
import time
import mimetypes
import re

import tornado.auth
import tornado.httpserver
import tornado.ioloop
import tornado.web
from tornado.options import define, options
from mako.template import Template

import waonline.core.waologgin as wl
from waonline.utils import custom_exception
from waonline.core.new_main import DatabaseManager
from waonline.core.users import UsersManager
from waonline.core import sessions
from waonline import conf
from waonline.core.ete_dumbobj import EteDumbObj

db_manager = None
utenti_manager = None
path_regexp = re.compile(r'^/app/(?P<application>\w+)/(?P<method>\w+)')

define("port", default=conf.DEFAULT_PORT, help="porta da utilizzare", type=int)
define("site_path", default=conf.DEFAULT_SITE,
        help="directory principale del sito")
define("url_db_utenti",
       default="%s/%s" % (conf.DEFAULT_DB_STR, conf.DEFAULT_DB_NAME),
       help="Stringa di connessione al db utenti.")
define("working_directory", conf.DEFAULT_WORKING,
        help="Cartella di lavoro")
define("data_path", conf.DEFAULT_DATA_PATH, help="Cartella Dati")
define("log_name", conf.DEFAULT_LOG_NAME, help="file di log")
define("daemon", default=1, help="0 per interattivo", type=int)
define("nose", default=0, help="1 per test", type=int)


sys.excepthook = custom_exception

#### Tornado Application
class Application(tornado.web.Application):
    """ Classe applicazione principale.  """
    def __init__(self, db_manager, utenti_manager, session_manager):
        handlers = [
            (r"/", IndexHandler),
            (r'^/app/utenti/login', LoginHandler),
            (r'^/app/(\w+)/(\w+)', AppHandler),
            (r"/templates/(.*)", TemplateHandler),
            (r"/ddoc", DDocHandler),
            (r"/dfile", DFileHandler),
            (r"/get", GetHandler),
            (r'/logout', LogoutHandler),
            (r"/SKIN/(.*)", SkinHandler),
            (r"/img/LogoStudio.png", LogoHandler),
            (r"/admin", AdminHandler),
            (r'/(.*)', tornado.web.StaticFileHandler, {"path": options.site_path}),
            (r'/js/sc/(.*)', tornado.web.StaticFileHandler, {"path": options.site_path}),
        ]
        settings = dict(
            cookie_secret="11oETzKXQAGaYdkL5gEmGeJJFuYh7EQnp2XdTP1o/Vo=",
            login_url="/app/utenti/login",
            debug=False,
            cookie_name="waonline",
            cookie_path="/"
        )
        tornado.web.Application.__init__(self, handlers, **settings)
        self.site_params = EteDumbObj(dict(
            nome = 'WAOnline',
            locale = 'utf-8',
            site_path = options.site_path,
            templace = conf.DEFAULT_TEMPLATE,
            skin = conf.SKIN_NAME))
        self.session_manager = session_manager
        self.utenti_manager = utenti_manager
        self.db_manager = db_manager

#### Request Handlers
class BaseHandler(tornado.web.RequestHandler):
    """Handler di base per le richieste """

    def prepare(self):
        self.dynamic_content = ('.htm', '.html', '.css')
        try:
            match = path_regexp.search(self.request.path)
            if match:
                self.nome_funzione = match.group('application')
                self.nome_task = match.group('method')
            else:
                self.nome_funzione = self.get_argument('type', None)
                self.nome_task = self.get_argument('task', None)
            meth = self.application.db_manager.get_method(self.nome_funzione, self.nome_task)
            self.cookie = self.get_secure_cookie(
                    self.application.settings['cookie_name'])
            self.session = self.application.session_manager.get_session(
                    self.cookie, lock=sessions.test_require_lock(meth))
            self.set_secure_cookie(self.application.settings['cookie_name'],
                    self.session.sid)
            self.startt = time.time()
            wl.debug("Func: %s - task: %s" % (self.nome_funzione, self.nome_task))
            self.request.utenti_manager = self.application.utenti_manager
            self.request.db_manager = self.application.db_manager
            #: Spostato il caricamento dell'utente nel prepare della richiesta
            if 'user' in self.session:
                self.user = self.request.user = \
                        self.request.utenti_manager.get_by_id(
                                self.session['user'])
                if self.user:
                    wl.info("Caricato l'utente %s" % self.user)
                else:
                    wl.info("Non esiste l'utente con ID:%s" % \
                            self.session['user'])
            else:
                self.user = self.request.user = None
            self.startt = time.time()
            if self.user:
                wl.debug("Prepare: user: %s %s" % (self.user, self.user.sessiondb))
            else:
                wl.debug("Prepare: Utente a None")
            self.template = self.session.data.get('tema', conf.DEFAULT_TEMPLATE)
            self.skin = self.session.data.get('skin', conf.SKIN_NAME)
            self.application.site_params.template = self.template
            self.application.site_params.skin = self.skin
            self.request.template = self.template
            self.request.skin = self.skin
            self.request.DEFAULT_TEMPLATE = conf.DEFAULT_TEMPLATE
            self.request.SKIN_NAME = conf.SKIN_NAME
        except:
            custom_exception()
            raise

    def get_current_user(self):
        return self.user

    def gestione_funzione(self):
        """gestione funzione"""
        self._gestione_funzione()

    @property
    def cod_studio(self):
        try:
            return self.user.studio_codice
        except:
            return None

    def _gestione_funzione(self):
        """callback per la gesione della funzione """
        wl.info("codice studio: %s" % self.cod_studio)
        try:
            risposta = self.request.db_manager.get_function(self.cod_studio,
                    self.nome_funzione, self.session, self.request)
            self.gestione_risposta(risposta)
        except:
            wl.debug("Ora parte custom_exception")
            custom_exception()
        # Separato, voglio che venga salvata comunque la sessione  FIXME
        try:
            self.session.save()
        except:
            custom_exception()
            raise
        self.finish()

    def gestione_risposta(self, risposta):
        """gestione risposta"""
        try:
            for (header, value) in risposta.header:
                self.set_header(header, value)
            if risposta.contenuto:
                self.write(risposta.contenuto)
            else:
                self.write('')
            wl.warn("Finito la funzione %s:%s (%2.3f)" % (
                self.nome_funzione,
                self.nome_task, ((time.time() - self.startt) * 1000)))
        except Exception, err:
            wl.debug("Errore in gestione risposta: %s" % err)

    def finish(self, chunk=None):
        tornado.web.RequestHandler.finish(self, chunk)
        if self.user and self.user.sessiondb:
            self.user.sessiondb.close()
            del self.user


class DynamicHandler(BaseHandler):
    """
    Classe per la gestione dei file template dinamici
    """
    def get_page(self, path, ext):
        """recupera la pagina"""
        try:
            is_dynamic = ext in self.dynamic_content
            self.set_header("Content-Type",
                    mimetypes.types_map.get(ext, "text/html"))
            if is_dynamic:
                return self.get_dynamic(path)
            else:
                data = self.get_file_data(path)
                return data
        except:
            custom_exception()
            raise

    def get_dynamic(self, path):
        """recupera dinamicamente"""
        try:
            return Template(filename=path).render(
                    sito=self.application.site_params)
        except:
            custom_exception()
            #return exceptions.html_error_template().render()
            raise

    def get_file_data(self, path):
        """recupera il file"""
        try:
            f = open(path, 'r')
            content = f.read()
            f.close()
            return content
        except:
            wl.errore("file: %s non trovato" % path)
            return ""


class IndexHandler(BaseHandler):
    """
    Classe per la gestione dell'index.
    """
    def get(self):
        self.redirect("templates/index.html")


class TemplateHandler(DynamicHandler):
    """classe per i template"""
    @tornado.web.authenticated
    def get(self, path):
        real_path = options.site_path + '/templates/' + \
        self.template +"/"+ path
        ext = os.path.splitext(path)[1]
        data = self.get_page(real_path, ext)
        self.write(data)
        self.finish()


class LoginHandler(DynamicHandler):
    """handler per il login"""
    def get(self):
        real_path = options.site_path + '/templates/' + \
                self.template + '/login.html'
        data = self.get_page(real_path, '.html')
        self.write(data)
        self.finish()

    def post(self):
        self.gestione_funzione()

    def gestione_funzione(self):
        """gestione funzione"""
        try:
            risposta = self.request.db_manager.call_app(None, 'utenti',
                    'login', self.session, self.request)
            self.gestione_risposta(risposta)
        except Exception, e:
            custom_exception()
            raise
            #raise tornado.web.HTTPError(500, str(e))
        try:
            self.session.save()
        except:
            custom_exception()
            raise
        self.finish()

    def gestione_risposta(self, risposta):
        """gestione della risposta"""
        wl.info("Gestion risposta %r" % risposta.header)
        for header, value in risposta.header:
            wl.info("Header: %s value; %s" % (header, value))
            self.set_header(header, value)
        if risposta.contenuto:
            self.write(risposta.contenuto)


class SkinHandler(DynamicHandler):
    """handler per le skins"""

    def get(self, skin_file):
        app_skin = '/js/sc/skins/' + self.skin + '/images'
        real_path = options.site_path + '/' + app_skin + '/' + skin_file
        wl.debug("REAL PATH: %s" % real_path)
        if not os.path.exists(real_path):
            wl.debug("Che pero non esiste")
            app_skin = '/js/sc/skins/' + conf.APPLICATION_SKIN + '/images'
            real_path = options.site_path + '/' + app_skin + '/' + skin_file

        data = self.get_page(real_path, '.gif')
        self.write(data)
        self.finish()


class LogoHandler(DynamicHandler):
    """Handler per il logo"""

    def get(self):
        real_path = ''
        if self.cod_studio:
            real_path = os.path.join(options.data_path, self.cod_studio,
            "LogoStudio.bmp")
        if not os.path.exists(real_path):
            real_path = os.path.join(options.site_path, "img/essepaghe1.png")
            wl.debug("real_path: %s site_path: %s" % (real_path, options.site_path))
        data = self.get_page(real_path, '.gif')
        self.write(data)
        self.finish()


class LogoutHandler(BaseHandler):
    """handler per il logout"""
    def get(self):
        # FIXME: lasciato commentato solo per questioni statistiche
        #self.session.delete()
        self.clear_all_cookies()
        self.redirect('/')

class DFileHandler(BaseHandler):
    """ Classe per la gestione delle get dei file"""

    def get(self):
        nome_file = self.get_argument('nome_file')
        nome_file = nome_file.split(".")[0]
        if self.application.session_manager.has_session(nome_file):
            cfile = open("/usr/local/wodata/moduli/%s.txt" % nome_file).read()
            self.set_header("Content-Disposition",
                    "inline; filename=ExportDef.txt")
            self.set_header("Content-Type", 'text/plain; charset="ISO-8859-1"')
            self.write(cfile)
        else:
            self.write('')
        self.finish()


class DDocHandler(BaseHandler):
    """ Classe per la gestione delle get dei file"""

    def get(self):
        nome_file = self.get_argument('nome_file')
        sid = self.get_argument('sid')
        ## FIXME: cosa serve?
        if self.application.session_manager.has_session(sid):
            cfile = open("/usr/local/wodata/moduli/segreteria/%s" \
                    % nome_file).read()
            self.set_header("Content-Disposition", "inline; filename=%s" \
                    % nome_file)
            self.set_header("Content-Type",
                    'application/msword; charset="ISO-8859-1"')
            self.write(cfile)
        else:
            self.write('')
        self.finish()

class AppHandler(BaseHandler):
    """ classe per gestire le applicazioni """

    @tornado.web.authenticated
    def get(self, application, method):
        self.gestione_funzione(application, method)

    @tornado.web.authenticated
    def post(self, application, method):
        self.gestione_funzione(application, method)

    def gestione_funzione(self, application, method):
        """callback per la gesione dell'applicazione """
        wl.info("APPLICAZIONE: %s/%s cod_studio: %s" % (application, method,
                self.cod_studio))
        try:
            risposta = self.request.db_manager.call_app(self.cod_studio,
                    application, method, self.session, self.request)
            self.gestione_risposta(risposta)
        except:
            wl.debug("Ora parte custom_exception")
            custom_exception()
        # Separato, voglio che venga salvata comunque la sessione  FIXME
        try:
            self.session.save()
        except:
            custom_exception()
            raise
        self.finish()


class GetHandler(BaseHandler):
    """ Classe per la gestione delle get """

    @tornado.web.authenticated
    def get(self):
        self.gestione_funzione()

    @tornado.web.authenticated
    def post(self):
        self.gestione_funzione()

class AdminHandler(BaseHandler):
    """Gestione delle request di admin"""

    def get(self):
        self.gestione_funzione()

    post = get


#### Program

if __name__ == "__main__":

    locale.setlocale(locale.LC_ALL, 'it_IT.utf8')
    tornado.options.parse_command_line()
    os.environ.update({"WAPATH": options.working_directory})
    os.sys.path.append(options.working_directory)
    if not 'In' in locals() and options.daemon:
        ctx = daemon.DaemonContext(working_directory=str(options.working_directory))
        ctx.open()
    wl.init_log()
    wl.PORTA = options.port
    wl.info("livello logging: %s" % options.logging)
    db_manager = DatabaseManager(db_url=conf.DEFAULT_DB_STR,
            db_base=conf.DEFAULT_DB_NAME, timeproxy=0)
    utenti_manager = UsersManager(db_manager)
    session_manager = sessions.SessionManager()
    if not 'In' in locals():
        app = Application(db_manager, utenti_manager, session_manager)
        http_server = tornado.httpserver.HTTPServer(app, xheaders=True)
        http_server.listen(options.port)
        tornado.ioloop.IOLoop.instance().start()
